@@ -84,9 +84,10 @@ typedef int socklen_t;
 #define USE_IPv6
 #define _CRT_SECURE_NO_DEPRECATE
 #define _CRT_NONSTDC_NO_DEPRECATE
+#define _CRT_NON_CONFORMING_SWPRINTFS
 #define HAVE_OSSL_ENGINE_H
 #define HAVE_OSSL_OCSP_H
-/* prevent including wincrypt.h, as it defines it's own OCSP_RESPONSE */
+/* prevent including wincrypt.h, as it defines its own OCSP_RESPONSE */
 #define __WINCRYPT_H__
 #endif
 
@@ -170,6 +171,11 @@ typedef int socklen_t;
 #define USE_LIBWRAP 1
 #endif
 
+/* systemd */
+#ifdef HAVE_SYSTEMD_SD_DAEMON_H
+#include <systemd/sd-daemon.h>
+#endif
+
 /* must be included before sys/stat.h for Ultrix */
 /* must be included before sys/socket.h for OpenBSD */
 #include <sys/types.h>   /* u_short, u_long */
@@ -440,11 +446,6 @@ extern char *sys_errlist[];
 #include <openssl/ocsp.h>
 #endif /* HAVE_OSSL_OCSP_H */
 
-#ifdef HAVE_OSSL_FIPS_H
-#include <openssl/fips.h>
-#include <openssl/fips_rand.h>
-#endif /* HAVE_OSSL_FIPS_H */
-
 #if OPENSSL_VERSION_NUMBER<0x0090800fL
 #define OPENSSL_NO_ECDH
 #endif /* OpenSSL version < 0.8.0 */
@@ -460,15 +461,6 @@ STACK_OF(SSL_COMP) *SSL_COMP_get_compression_methods(void);
 
 /**************************************** other defines */
 
-/* change all non-printable characters to '.' */
-#define safestring(s) \
-    do {unsigned char *p; for(p=(unsigned char *)(s); *p; p++) \
-        if(!isprint((int)*p)) *p='.';} while(0)
-/* change all unsafe characters to '.' */
-#define safename(s) \
-    do {unsigned char *p; for(p=(s); *p; p++) \
-        if(!isalnum((int)*p)) *p='.';} while(0)
-
 /* always use IPv4 defaults! */
 #define DEFAULT_LOOPBACK "127.0.0.1"
 #define DEFAULT_ANY "0.0.0.0"
