@@ -38,30 +38,117 @@
 #include "common.h"
 #include "prototypes.h"
 
-#if !defined(OPENSSL_NO_TLS1)
-#define DEFAULT_SSLVER_CLIENT "TLSv1"
-#elif !defined(OPENSSL_NO_SSL3)
-#define DEFAULT_SSLVER_CLIENT "SSLv3"
-#elif !defined(OPENSSL_NO_SSL2)
-#define DEFAULT_SSLVER_CLIENT "SSLv2"
-#else /* OPENSSL_NO_TLS1, OPENSSL_NO_SSL3, OPENSSL_NO_SSL2 */
-#error No supported SSL methods found
-#endif /* OPENSSL_NO_TLS1, OPENSSL_NO_SSL3, OPENSSL_NO_SSL2 */
-#define DEFAULT_SSLVER_SERVER "all"
-
 #if defined(_WIN32_WCE) && !defined(CONFDIR)
 #define CONFDIR "\\stunnel"
 #endif
 
 #define CONFLINELEN (16*1024)
 
+typedef enum {
+    CMD_BEGIN,      /* initialize defaults */
+    CMD_EXEC,       /* process command */
+    CMD_END,        /* end of section */
+    CMD_FREE,       /* TODO: deallocate memory */
+    CMD_DEFAULT,    /* print default value */
+    CMD_HELP        /* print help */
+} CMD;
+
+NOEXPORT char *parse_global_option(CMD, char *, char *);
+NOEXPORT char *parse_service_option(CMD, SERVICE_OPTIONS *, char *, char *);
+
 #ifndef OPENSSL_NO_TLSEXT
 NOEXPORT char *init_sni(SERVICE_OPTIONS *);
 #endif
 
 NOEXPORT int parse_debug_level(char *);
 
+typedef struct {
+    char *name;
+    long value;
+} SSL_OPTION;
+
+static const SSL_OPTION ssl_opts[] = {
+    {"MICROSOFT_SESS_ID_BUG", SSL_OP_MICROSOFT_SESS_ID_BUG},
+    {"NETSCAPE_CHALLENGE_BUG", SSL_OP_NETSCAPE_CHALLENGE_BUG},
+#ifdef SSL_OP_LEGACY_SERVER_CONNECT
+    {"LEGACY_SERVER_CONNECT", SSL_OP_LEGACY_SERVER_CONNECT},
+#endif
+    {"NETSCAPE_REUSE_CIPHER_CHANGE_BUG",
+        SSL_OP_NETSCAPE_REUSE_CIPHER_CHANGE_BUG},
+#ifdef SSL_OP_TLSEXT_PADDING
+    {"TLSEXT_PADDING", SSL_OP_TLSEXT_PADDING},
+#endif
+    {"MICROSOFT_BIG_SSLV3_BUFFER", SSL_OP_MICROSOFT_BIG_SSLV3_BUFFER},
+#ifdef SSL_OP_SAFARI_ECDHE_ECDSA_BUG
+    {"SAFARI_ECDHE_ECDSA_BUG", SSL_OP_SAFARI_ECDHE_ECDSA_BUG},
+#endif
+    {"SSLEAY_080_CLIENT_DH_BUG", SSL_OP_SSLEAY_080_CLIENT_DH_BUG},
+    {"TLS_D5_BUG", SSL_OP_TLS_D5_BUG},
+    {"TLS_BLOCK_PADDING_BUG", SSL_OP_TLS_BLOCK_PADDING_BUG},
+#ifdef SSL_OP_MSIE_SSLV2_RSA_PADDING
+    {"MSIE_SSLV2_RSA_PADDING", SSL_OP_MSIE_SSLV2_RSA_PADDING},
+#endif
+    {"SSLREF2_REUSE_CERT_TYPE_BUG", SSL_OP_SSLREF2_REUSE_CERT_TYPE_BUG},
+#ifdef SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
+    {"DONT_INSERT_EMPTY_FRAGMENTS", SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS},
+#endif
+    {"ALL", SSL_OP_ALL},
+#ifdef SSL_OP_NO_QUERY_MTU
+    {"NO_QUERY_MTU", SSL_OP_NO_QUERY_MTU},
+#endif
+#ifdef SSL_OP_COOKIE_EXCHANGE
+    {"COOKIE_EXCHANGE", SSL_OP_COOKIE_EXCHANGE},
+#endif
+#ifdef SSL_OP_NO_TICKET
+    {"NO_TICKET", SSL_OP_NO_TICKET},
+#endif
+#ifdef SSL_OP_CISCO_ANYCONNECT
+    {"CISCO_ANYCONNECT", SSL_OP_CISCO_ANYCONNECT},
+#endif
+#ifdef SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION
+    {"NO_SESSION_RESUMPTION_ON_RENEGOTIATION",
+        SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION},
+#endif
+#ifdef SSL_OP_NO_COMPRESSION
+    {"NO_COMPRESSION", SSL_OP_NO_COMPRESSION},
+#endif
+#ifdef SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
+    {"ALLOW_UNSAFE_LEGACY_RENEGOTIATION",
+        SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION},
+#endif
+#ifdef SSL_OP_SINGLE_ECDH_USE
+    {"SINGLE_ECDH_USE", SSL_OP_SINGLE_ECDH_USE},
+#endif
+    {"SINGLE_DH_USE", SSL_OP_SINGLE_DH_USE},
+    {"EPHEMERAL_RSA", SSL_OP_EPHEMERAL_RSA},
+#ifdef SSL_OP_CIPHER_SERVER_PREFERENCE
+    {"CIPHER_SERVER_PREFERENCE", SSL_OP_CIPHER_SERVER_PREFERENCE},
+#endif
+    {"TLS_ROLLBACK_BUG", SSL_OP_TLS_ROLLBACK_BUG},
+    {"NO_SSLv2", SSL_OP_NO_SSLv2},
+    {"NO_SSLv3", SSL_OP_NO_SSLv3},
+    {"NO_TLSv1", SSL_OP_NO_TLSv1},
+#ifdef SSL_OP_NO_TLSv1_1
+    {"NO_TLSv1.1", SSL_OP_NO_TLSv1_1},
+#endif
+#ifdef SSL_OP_NO_TLSv1_2
+    {"NO_TLSv1.2", SSL_OP_NO_TLSv1_2},
+#endif
+    {"PKCS1_CHECK_1", SSL_OP_PKCS1_CHECK_1},
+    {"PKCS1_CHECK_2", SSL_OP_PKCS1_CHECK_2},
+    {"NETSCAPE_CA_DN_BUG", SSL_OP_NETSCAPE_CA_DN_BUG},
+#ifdef SSL_OP_NON_EXPORT_FIRST
+    {"NON_EXPORT_FIRST", SSL_OP_NON_EXPORT_FIRST},
+#endif
+    {"NETSCAPE_DEMO_CIPHER_CHANGE_BUG", SSL_OP_NETSCAPE_DEMO_CIPHER_CHANGE_BUG},
+#ifdef SSL_OP_CRYPTOPRO_TLSEXT_BUG
+    {"CRYPTOPRO_TLSEXT_BUG", SSL_OP_CRYPTOPRO_TLSEXT_BUG},
+#endif
+    {NULL, 0}
+};
+
 NOEXPORT int parse_ssl_option(char *);
+NOEXPORT void print_ssl_options(void);
 
 NOEXPORT int print_socket_options(void);
 NOEXPORT char *print_option(int, OPT_UNION *);
@@ -89,7 +176,16 @@ NOEXPORT void print_syntax(void);
 NOEXPORT char **argalloc(char *);
 #endif
 
-char *configuration_file;
+char *configuration_file=
+#ifdef CONFDIR
+            CONFDIR
+#ifdef USE_WIN32
+            "\\"
+#else
+            "/"
+#endif
+#endif
+            "stunnel.conf";
 
 GLOBAL_OPTIONS global_options;
 SERVICE_OPTIONS service_options;
@@ -97,21 +193,191 @@ SERVICE_OPTIONS service_options;
 static GLOBAL_OPTIONS new_global_options;
 static SERVICE_OPTIONS new_service_options;
 
-typedef enum {
-    CMD_BEGIN,      /* initialize defaults */
-    CMD_EXEC,       /* process command */
-    CMD_END,        /* end of section */
-    CMD_FREE,       /* TODO: deallocate memory */
-    CMD_DEFAULT,    /* print default value */
-    CMD_HELP        /* print help */
-} CMD;
-
 static char *option_not_found=
     "Specified option name is not valid here";
 
 static char *stunnel_cipher_list=
     "HIGH:MEDIUM:+3DES:+DH:!aNULL:!SSLv2";
 
+/**************************************** parse commandline parameters */
+
+int options_cmdline(char *name, char *parameter) {
+    CONF_TYPE type=CONF_FILE;
+
+#ifdef USE_WIN32
+    (void)parameter; /* skip warning about unused parameter */
+#endif
+    if(!name) {
+        /* leave the previous value of configuration_file */
+    } else if(!strcasecmp(name, "-help")) {
+        parse_global_option(CMD_HELP, NULL, NULL);
+        parse_service_option(CMD_HELP, NULL, NULL, NULL);
+        log_flush(LOG_MODE_INFO);
+        return 1;
+    } else if(!strcasecmp(name, "-version")) {
+        parse_global_option(CMD_DEFAULT, NULL, NULL);
+        parse_service_option(CMD_DEFAULT, NULL, NULL, NULL);
+        log_flush(LOG_MODE_INFO);
+        return 1;
+    } else if(!strcasecmp(name, "-sockets")) {
+        print_socket_options();
+        log_flush(LOG_MODE_INFO);
+        return 1;
+    } else if(!strcasecmp(name, "-options")) {
+        print_ssl_options();
+        log_flush(LOG_MODE_INFO);
+        return 1;
+    } else
+#ifndef USE_WIN32
+    if(!strcasecmp(name, "-fd")) {
+        if(!parameter) {
+            s_log(LOG_ERR, "No file descriptor specified");
+            print_syntax();
+            return 1;
+        }
+        configuration_file=parameter;
+        type=CONF_FD;
+    } else
+#endif
+        configuration_file=name;
+    configuration_file=str_dup(configuration_file);
+    str_detach(configuration_file); /* do not track this allocation */
+
+    return options_parse(type);
+}
+
+/**************************************** parse configuration file */
+
+int options_parse(CONF_TYPE type) {
+    DISK_FILE *df;
+    char line_text[CONFLINELEN], *errstr;
+    char config_line[CONFLINELEN], *config_opt, *config_arg;
+    int line_number, i;
+    SERVICE_OPTIONS *section, *new_section;
+#ifndef USE_WIN32
+    int fd;
+    char *tmpstr;
+#endif
+
+    s_log(LOG_NOTICE, "Reading configuration from %s %s",
+        type==CONF_FD ? "descriptor" : "file", configuration_file);
+#ifndef USE_WIN32
+    if(type==CONF_FD) { /* file descriptor */
+        fd=strtol(configuration_file, &tmpstr, 10);
+        if(tmpstr==configuration_file || *tmpstr) { /* not a number */
+            s_log(LOG_ERR, "Invalid file descriptor number");
+            print_syntax();
+            return 1;
+        }
+        df=file_fdopen(fd);
+    } else
+#endif
+        df=file_open(configuration_file, FILE_MODE_READ);
+    if(!df) {
+        s_log(LOG_ERR, "Cannot open configuration file");
+        if(type!=CONF_RELOAD)
+            print_syntax();
+        return 1;
+    }
+
+    options_defaults();
+    section=&new_service_options;
+    line_number=0;
+    while(file_getline(df, line_text, CONFLINELEN)>=0) {
+        memcpy(config_line, line_text, CONFLINELEN);
+        ++line_number;
+        config_opt=config_line;
+        while(isspace((unsigned char)*config_opt))
+            ++config_opt; /* remove initial whitespaces */
+        for(i=strlen(config_opt)-1; i>=0 && isspace((unsigned char)config_opt[i]); --i)
+            config_opt[i]='\0'; /* remove trailing whitespaces */
+        if(config_opt[0]=='\0' || config_opt[0]=='#' || config_opt[0]==';') /* empty or comment */
+            continue;
+        if(config_opt[0]=='[' && config_opt[strlen(config_opt)-1]==']') { /* new section */
+            if(!new_service_options.next) {
+                errstr=parse_global_option(CMD_END, NULL, NULL);
+                if(errstr) {
+                    s_log(LOG_ERR, "Line %d: \"%s\": %s",
+                        line_number, line_text, errstr);
+                    file_close(df);
+                    return 1;
+                }
+            }
+            ++config_opt;
+            config_opt[strlen(config_opt)-1]='\0';
+            new_section=str_alloc(sizeof(SERVICE_OPTIONS));
+            memcpy(new_section, &new_service_options, sizeof(SERVICE_OPTIONS));
+            new_section->servname=str_dup(config_opt);
+            new_section->session=NULL;
+            new_section->next=NULL;
+            section->next=new_section;
+            section=new_section;
+            continue;
+        }
+        config_arg=strchr(config_line, '=');
+        if(!config_arg) {
+            s_log(LOG_ERR, "Line %d: \"%s\": No '=' found", line_number, line_text);
+            file_close(df);
+            return 1;
+        }
+        *config_arg++='\0'; /* split into option name and argument value */
+        for(i=strlen(config_opt)-1; i>=0 && isspace((unsigned char)config_opt[i]); --i)
+            config_opt[i]='\0'; /* remove trailing whitespaces */
+        while(isspace((unsigned char)*config_arg))
+            ++config_arg; /* remove initial whitespaces */
+        errstr=parse_service_option(CMD_EXEC, section, config_opt, config_arg);
+        if(!new_service_options.next && errstr==option_not_found)
+            errstr=parse_global_option(CMD_EXEC, config_opt, config_arg);
+        if(errstr) {
+            s_log(LOG_ERR, "Line %d: \"%s\": %s", line_number, line_text, errstr);
+            file_close(df);
+            return 1;
+        }
+    }
+    file_close(df);
+
+    if(new_service_options.next) { /* daemon mode: initialize sections */
+        for(section=new_service_options.next; section; section=section->next) {
+            s_log(LOG_INFO, "Initializing service [%s]", section->servname);
+            errstr=parse_service_option(CMD_END, section, NULL, NULL);
+            if(errstr)
+                break;
+        }
+    } else { /* inetd mode: need to initialize global options */
+        errstr=parse_global_option(CMD_END, NULL, NULL);
+        if(errstr) {
+            s_log(LOG_ERR, "Global options: %s", errstr);
+            return 1;
+        }
+        s_log(LOG_INFO, "Initializing inetd mode configuration");
+        section=&new_service_options;
+        errstr=parse_service_option(CMD_END, section, NULL, NULL);
+    }
+    if(errstr) {
+        s_log(LOG_ERR, "Service [%s]: %s", section->servname, errstr);
+        return 1;
+    }
+
+    s_log(LOG_NOTICE, "Configuration successful");
+    return 0;
+}
+
+void options_defaults() {
+    /* initialize globals *before* opening the config file */
+    memset(&new_global_options, 0, sizeof(GLOBAL_OPTIONS)); /* reset global options */
+    memset(&new_service_options, 0, sizeof(SERVICE_OPTIONS)); /* reset local options */
+    new_service_options.next=NULL;
+    parse_global_option(CMD_BEGIN, NULL, NULL);
+    parse_service_option(CMD_BEGIN, &new_service_options, NULL, NULL);
+}
+
+void options_apply() { /* apply default/validated configuration */
+    /* FIXME: this operation may be unsafe, as client() threads use it */
+    memcpy(&global_options, &new_global_options, sizeof(GLOBAL_OPTIONS));
+    /* service_options are used for inetd mode and to enumerate services */
+    memcpy(&service_options, &new_service_options, sizeof(SERVICE_OPTIONS));
+}
+
 /**************************************** global options */
 
 NOEXPORT char *parse_global_option(CMD cmd, char *opt, char *arg) {
@@ -765,44 +1031,7 @@ NOEXPORT char *parse_global_option(CMD cmd, char *opt, char *arg) {
         /* FIPS needs to be initialized as early as possible */
         if(ssl_configure(&new_global_options)) /* configure global SSL settings */
             return "Failed to initialize SSL";
-
-        /* prepare default SSL methods */
-#ifdef USE_FIPS
-        if(new_global_options.option.fips) {
-            if(!new_service_options.cipher_list)
-                new_service_options.cipher_list="FIPS";
-            if(!new_service_options.client_method)
-                new_service_options.client_method=
-                    (SSL_METHOD *)TLSv1_client_method();
-            if(!new_service_options.server_method)
-                new_service_options.server_method=
-                    (SSL_METHOD *)TLSv1_server_method();
-        } else {
-#endif /* USE_FIPS */
-            if(!new_service_options.cipher_list)
-                new_service_options.cipher_list=stunnel_cipher_list;
-            if(!new_service_options.client_method)
-#if !defined(OPENSSL_NO_TLS1)
-                new_service_options.client_method=
-                    (SSL_METHOD *)TLSv1_client_method();
-#elif !defined(OPENSSL_NO_SSL3)
-                new_service_options.client_method=
-                    (SSL_METHOD *)SSLv3_client_method();
-#elif !defined(OPENSSL_NO_SSL2)
-                new_service_options.client_method=
-                    (SSL_METHOD *)SSLv2_client_method();
-#else /* OPENSSL_NO_TLS1, OPENSSL_NO_SSL3, OPENSSL_NO_SSL2 */
-#error No supported SSL methods found
-#endif /* OPENSSL_NO_TLS1, OPENSSL_NO_SSL3, OPENSSL_NO_SSL2 */
-            /* SSLv23_server_method() is an always available catch-all */
-            if(!new_service_options.server_method)
-                new_service_options.server_method=
-                    (SSL_METHOD *)SSLv23_server_method();
-#ifdef USE_FIPS
-        }
-#endif /* USE_FIPS */
     }
-
     return NULL; /* OK */
 }
 
@@ -949,6 +1178,17 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
         section->cipher_list=str_dup(arg);
         return NULL; /* OK */
     case CMD_END:
+#ifdef USE_FIPS
+        if(new_global_options.option.fips) {
+            if(!new_service_options.cipher_list)
+                new_service_options.cipher_list="FIPS";
+        } else
+#endif /* USE_FIPS */
+        {
+            if(!new_service_options.cipher_list)
+                new_service_options.cipher_list=stunnel_cipher_list;
+        }
+
         break;
     case CMD_FREE:
         break;
@@ -1427,15 +1667,23 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     /* options */
     switch(cmd) {
     case CMD_BEGIN:
-        section->ssl_options=0;
+        section->ssl_options_set|=SSL_OP_NO_SSLv2|SSL_OP_NO_SSLv3;
+        section->ssl_options_clear=0;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "options"))
             break;
-        tmpnum=parse_ssl_option(arg);
-        if(!tmpnum)
-            return "Illegal SSL option";
-        section->ssl_options|=tmpnum;
+        if(*arg=='-') {
+            tmpnum=parse_ssl_option(arg+1);
+            if(!tmpnum)
+                return "Illegal SSL option";
+            section->ssl_options_clear|=tmpnum;
+        } else {
+            tmpnum=parse_ssl_option(arg);
+            if(!tmpnum)
+                return "Illegal SSL option";
+            section->ssl_options_set|=tmpnum;
+        }
         return NULL; /* OK */
     case CMD_END:
         break;
@@ -1770,7 +2018,7 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
 #ifdef SSL_OP_NO_TICKET
         /* disable RFC4507 support introduced in OpenSSL 0.9.8f */
         /* this prevents session callbacks from beeing executed */
-        section->ssl_options|=SSL_OP_NO_TICKET;
+        section->ssl_options_set|=SSL_OP_NO_TICKET;
 #endif
         if(!name2addr(&section->sessiond_addr, arg, DEFAULT_LOOPBACK))
             return "Failed to resolve sessiond server address";
@@ -1821,8 +2069,8 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     /* sslVersion */
     switch(cmd) {
     case CMD_BEGIN:
-        section->client_method=NULL;
-        section->server_method=NULL;
+        section->client_method=(SSL_METHOD *)SSLv23_client_method();
+        section->server_method=(SSL_METHOD *)SSLv23_server_method();;
         break;
     case CMD_EXEC:
         if(strcasecmp(opt, "sslVersion"))
@@ -1872,14 +2120,12 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
 #ifdef USE_FIPS
         if(new_global_options.option.fips) {
             if(section->option.client) {
-                if(section->client_method!=(SSL_METHOD *)TLSv1_client_method() &&
-                        section->client_method!=(SSL_METHOD *)TLSv1_1_client_method() &&
-                        section->client_method!=(SSL_METHOD *)TLSv1_2_client_method())
+                if(section->client_method==(SSL_METHOD *)SSLv2_client_method() ||
+                        section->client_method==(SSL_METHOD *)SSLv3_client_method())
                     return "FIPS mode requires sslVersion to be TLSv1 or later";
             } else {
-                if(section->server_method!=(SSL_METHOD *)TLSv1_server_method() &&
-                        section->server_method!=(SSL_METHOD *)TLSv1_1_server_method() &&
-                        section->server_method!=(SSL_METHOD *)TLSv1_2_server_method())
+                if(section->server_method==(SSL_METHOD *)SSLv2_server_method() ||
+                        section->server_method==(SSL_METHOD *)SSLv3_server_method())
                     return "FIPS mode requires sslVersion to be TLSv1 or later";
             }
         }
@@ -1888,16 +2134,6 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     case CMD_FREE:
         break;
     case CMD_DEFAULT:
-#ifdef USE_FIPS
-        s_log(LOG_NOTICE, "%-22s = TLSv1 (with \"fips = yes\")",
-                "sslVersion");
-        s_log(LOG_NOTICE, "%-22s = " DEFAULT_SSLVER_CLIENT " for client, "
-                DEFAULT_SSLVER_SERVER " for server (with \"fips = no\")",
-                "sslVersion");
-#else
-        s_log(LOG_NOTICE, "%-22s = " DEFAULT_SSLVER_CLIENT " for client, "
-                DEFAULT_SSLVER_SERVER " for server", "sslVersion");
-#endif
         break;
     case CMD_HELP:
         s_log(LOG_NOTICE, "%-22s = all|SSLv2|SSLv3|TLSv1"
@@ -2133,194 +2369,6 @@ NOEXPORT char *parse_service_option(CMD cmd, SERVICE_OPTIONS *section,
     return NULL; /* OK */
 }
 
-/**************************************** parse commandline parameters */
-
-int parse_commandline(char *name, char *parameter) {
-    CONF_TYPE type=CONF_FILE;
-
-#ifdef USE_WIN32
-    (void)parameter; /* skip warning about unused parameter */
-#endif
-    if(!name) {
-        configuration_file=
-#ifdef CONFDIR
-            CONFDIR
-#ifdef USE_WIN32
-            "\\"
-#else
-            "/"
-#endif
-#endif
-            "stunnel.conf";
-    } else if(!strcasecmp(name, "-help")) {
-        parse_global_option(CMD_HELP, NULL, NULL);
-        parse_service_option(CMD_HELP, NULL, NULL, NULL);
-        log_flush(LOG_MODE_INFO);
-        return 1;
-    } else if(!strcasecmp(name, "-version")) {
-        parse_global_option(CMD_DEFAULT, NULL, NULL);
-        parse_service_option(CMD_DEFAULT, NULL, NULL, NULL);
-        log_flush(LOG_MODE_INFO);
-        return 1;
-    } else if(!strcasecmp(name, "-sockets")) {
-        print_socket_options();
-        log_flush(LOG_MODE_INFO);
-        return 1;
-    } else
-#ifndef USE_WIN32
-    if(!strcasecmp(name, "-fd")) {
-        if(!parameter) {
-            s_log(LOG_ERR, "No file descriptor specified");
-            print_syntax();
-            return 1;
-        }
-        configuration_file=parameter;
-        type=CONF_FD;
-    } else
-#endif
-        configuration_file=name;
-    configuration_file=str_dup(configuration_file);
-    str_detach(configuration_file); /* do not track this allocation */
-
-    if(parse_conf(type))
-        return 1;
-    apply_conf();
-    return 0;
-}
-
-/**************************************** parse configuration file */
-
-int parse_conf(CONF_TYPE type) {
-    DISK_FILE *df;
-    char line_text[CONFLINELEN], *errstr;
-    char config_line[CONFLINELEN], *config_opt, *config_arg;
-    int line_number, i;
-    SERVICE_OPTIONS *section, *new_section;
-#ifndef USE_WIN32
-    int fd;
-    char *tmpstr;
-#endif
-
-    /* initialize globals *before* opening the config file */
-    memset(&new_global_options, 0, sizeof(GLOBAL_OPTIONS)); /* reset global options */
-    memset(&new_service_options, 0, sizeof(SERVICE_OPTIONS)); /* reset local options */
-    new_service_options.next=NULL;
-    section=&new_service_options;
-    parse_global_option(CMD_BEGIN, NULL, NULL);
-    parse_service_option(CMD_BEGIN, section, NULL, NULL);
-    if(type!=CONF_RELOAD) { /* provide defaults for ui module */
-        memcpy(&global_options, &new_global_options, sizeof(GLOBAL_OPTIONS));
-        memcpy(&service_options, &new_service_options, sizeof(SERVICE_OPTIONS));
-    }
-
-    s_log(LOG_NOTICE, "Reading configuration from %s %s",
-        type==CONF_FD ? "descriptor" : "file", configuration_file);
-#ifndef USE_WIN32
-    if(type==CONF_FD) { /* file descriptor */
-        fd=strtol(configuration_file, &tmpstr, 10);
-        if(tmpstr==configuration_file || *tmpstr) { /* not a number */
-            s_log(LOG_ERR, "Invalid file descriptor number");
-            print_syntax();
-            return 1;
-        }
-        df=file_fdopen(fd);
-    } else
-#endif
-        df=file_open(configuration_file, FILE_MODE_READ);
-    if(!df) {
-        s_log(LOG_ERR, "Cannot open configuration file");
-        if(type!=CONF_RELOAD)
-            print_syntax();
-        return 1;
-    }
-
-    line_number=0;
-    while(file_getline(df, line_text, CONFLINELEN)>=0) {
-        memcpy(config_line, line_text, CONFLINELEN);
-        ++line_number;
-        config_opt=config_line;
-        while(isspace((unsigned char)*config_opt))
-            ++config_opt; /* remove initial whitespaces */
-        for(i=strlen(config_opt)-1; i>=0 && isspace((unsigned char)config_opt[i]); --i)
-            config_opt[i]='\0'; /* remove trailing whitespaces */
-        if(config_opt[0]=='\0' || config_opt[0]=='#' || config_opt[0]==';') /* empty or comment */
-            continue;
-        if(config_opt[0]=='[' && config_opt[strlen(config_opt)-1]==']') { /* new section */
-            if(!new_service_options.next) {
-                errstr=parse_global_option(CMD_END, NULL, NULL);
-                if(errstr) {
-                    s_log(LOG_ERR, "Line %d: \"%s\": %s", line_number, line_text, errstr);
-                    file_close(df);
-                    return 1;
-                }
-            }
-            ++config_opt;
-            config_opt[strlen(config_opt)-1]='\0';
-            new_section=str_alloc(sizeof(SERVICE_OPTIONS));
-            memcpy(new_section, &new_service_options, sizeof(SERVICE_OPTIONS));
-            new_section->servname=str_dup(config_opt);
-            new_section->session=NULL;
-            new_section->next=NULL;
-            section->next=new_section;
-            section=new_section;
-            continue;
-        }
-        config_arg=strchr(config_line, '=');
-        if(!config_arg) {
-            s_log(LOG_ERR, "Line %d: \"%s\": No '=' found", line_number, line_text);
-            file_close(df);
-            return 1;
-        }
-        *config_arg++='\0'; /* split into option name and argument value */
-        for(i=strlen(config_opt)-1; i>=0 && isspace((unsigned char)config_opt[i]); --i)
-            config_opt[i]='\0'; /* remove trailing whitespaces */
-        while(isspace((unsigned char)*config_arg))
-            ++config_arg; /* remove initial whitespaces */
-        errstr=parse_service_option(CMD_EXEC, section, config_opt, config_arg);
-        if(!new_service_options.next && errstr==option_not_found)
-            errstr=parse_global_option(CMD_EXEC, config_opt, config_arg);
-        if(errstr) {
-            s_log(LOG_ERR, "Line %d: \"%s\": %s", line_number, line_text, errstr);
-            file_close(df);
-            return 1;
-        }
-    }
-    file_close(df);
-
-    if(new_service_options.next) { /* daemon mode: initialize sections */
-        for(section=new_service_options.next; section; section=section->next) {
-            s_log(LOG_INFO, "Initializing service [%s]", section->servname);
-            errstr=parse_service_option(CMD_END, section, NULL, NULL);
-            if(errstr)
-                break;
-        }
-    } else { /* inetd mode: need to initialize global options */
-        errstr=parse_global_option(CMD_END, NULL, NULL);
-        if(errstr) {
-            s_log(LOG_ERR, "Global options: %s", errstr);
-            return 1;
-        }
-        s_log(LOG_INFO, "Initializing inetd mode configuration");
-        section=&new_service_options;
-        errstr=parse_service_option(CMD_END, section, NULL, NULL);
-    }
-    if(errstr) {
-        s_log(LOG_ERR, "Service [%s]: %s", section->servname, errstr);
-        return 1;
-    }
-
-    s_log(LOG_NOTICE, "Configuration successful");
-    return 0;
-}
-
-void apply_conf() { /* can be used once the configuration was validated */
-    /* FIXME: this operation may be unsafe, as client() threads use it */
-    memcpy(&global_options, &new_global_options, sizeof(GLOBAL_OPTIONS));
-    /* service_options are used for inetd mode and to enumerate services */
-    memcpy(&service_options, &new_service_options, sizeof(SERVICE_OPTIONS));
-    ui_new_config();
-}
-
 /**************************************** validate and initialize configuration */
 
 #ifndef OPENSSL_NO_TLSEXT
@@ -2328,7 +2376,7 @@ NOEXPORT char *init_sni(SERVICE_OPTIONS *section) {
     char *tmpstr;
     SERVICE_OPTIONS *tmpsrv;
 
-    /* server mode: update servername_list based on SNI option */
+    /* server mode: update servername_list based on the SNI option */
     if(!section->option.client && section->sni) {
         tmpstr=strchr(section->sni, ':');
         if(!tmpstr)
@@ -2348,7 +2396,8 @@ NOEXPORT char *init_sni(SERVICE_OPTIONS *section) {
             tmpsrv->servername_list_head=
                 tmpsrv->servername_list_tail=
                 str_alloc(sizeof(SERVERNAME_LIST));
-            tmpsrv->ssl_options|=SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
+            tmpsrv->ssl_options_set|=
+                SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
         }
         tmpsrv->servername_list_tail->servername=str_dup(tmpstr);
         tmpsrv->servername_list_tail->opt=section;
@@ -2356,7 +2405,8 @@ NOEXPORT char *init_sni(SERVICE_OPTIONS *section) {
         section->option.sni=1;
         /* always negotiate a new session on renegotiation, as the SSL
          * context settings (including access control) may be different */
-        section->ssl_options|=SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
+        section->ssl_options_set|=
+            SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION;
     }
 
     /* client mode: setup SNI default based on 'protocolHost' and 'connect' options */
@@ -2465,87 +2515,23 @@ NOEXPORT int parse_debug_level(char *arg) {
 /**************************************** SSL options */
 
 NOEXPORT int parse_ssl_option(char *arg) {
-    struct {
-        char *name;
-        long value;
-    } ssl_opts[] = {
-        {"MICROSOFT_SESS_ID_BUG", SSL_OP_MICROSOFT_SESS_ID_BUG},
-        {"NETSCAPE_CHALLENGE_BUG", SSL_OP_NETSCAPE_CHALLENGE_BUG},
-#ifdef SSL_OP_LEGACY_SERVER_CONNECT
-        {"LEGACY_SERVER_CONNECT", SSL_OP_LEGACY_SERVER_CONNECT},
-#endif
-        {"NETSCAPE_REUSE_CIPHER_CHANGE_BUG",
-            SSL_OP_NETSCAPE_REUSE_CIPHER_CHANGE_BUG},
-        {"SSLREF2_REUSE_CERT_TYPE_BUG", SSL_OP_SSLREF2_REUSE_CERT_TYPE_BUG},
-        {"MICROSOFT_BIG_SSLV3_BUFFER", SSL_OP_MICROSOFT_BIG_SSLV3_BUFFER},
-#ifdef SSL_OP_MSIE_SSLV2_RSA_PADDING
-        {"MSIE_SSLV2_RSA_PADDING", SSL_OP_MSIE_SSLV2_RSA_PADDING},
-#endif
-        {"SSLEAY_080_CLIENT_DH_BUG", SSL_OP_SSLEAY_080_CLIENT_DH_BUG},
-        {"TLS_D5_BUG", SSL_OP_TLS_D5_BUG},
-        {"TLS_BLOCK_PADDING_BUG", SSL_OP_TLS_BLOCK_PADDING_BUG},
-#ifdef SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS
-        {"DONT_INSERT_EMPTY_FRAGMENTS", SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS},
-#endif
-        {"ALL", SSL_OP_ALL},
-#ifdef SSL_OP_NO_QUERY_MTU
-        {"NO_QUERY_MTU", SSL_OP_NO_QUERY_MTU},
-#endif
-#ifdef SSL_OP_COOKIE_EXCHANGE
-        {"COOKIE_EXCHANGE", SSL_OP_COOKIE_EXCHANGE},
-#endif
-#ifdef SSL_OP_NO_TICKET
-        {"NO_TICKET", SSL_OP_NO_TICKET},
-#endif
-#ifdef SSL_OP_CISCO_ANYCONNECT
-        {"CISCO_ANYCONNECT", SSL_OP_CISCO_ANYCONNECT},
-#endif
-#ifdef SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION
-        {"NO_SESSION_RESUMPTION_ON_RENEGOTIATION",
-            SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION},
-#endif
-#ifdef SSL_OP_NO_COMPRESSION
-        {"NO_COMPRESSION", SSL_OP_NO_COMPRESSION},
-#endif
-#ifdef SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION
-        {"ALLOW_UNSAFE_LEGACY_RENEGOTIATION",
-            SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION},
-#endif
-#ifdef SSL_OP_SINGLE_ECDH_USE
-        {"SINGLE_ECDH_USE", SSL_OP_SINGLE_ECDH_USE},
-#endif
-        {"SINGLE_DH_USE", SSL_OP_SINGLE_DH_USE},
-        {"EPHEMERAL_RSA", SSL_OP_EPHEMERAL_RSA},
-#ifdef SSL_OP_CIPHER_SERVER_PREFERENCE
-        {"CIPHER_SERVER_PREFERENCE", SSL_OP_CIPHER_SERVER_PREFERENCE},
-#endif
-        {"TLS_ROLLBACK_BUG", SSL_OP_TLS_ROLLBACK_BUG},
-        {"NO_SSLv2", SSL_OP_NO_SSLv2},
-        {"NO_SSLv3", SSL_OP_NO_SSLv3},
-        {"NO_TLSv1", SSL_OP_NO_TLSv1},
-#ifdef SSL_OP_NO_TLSv1_1
-        {"NO_TLSv1.1", SSL_OP_NO_TLSv1_1},
-#endif
-#ifdef SSL_OP_NO_TLSv1_2
-        {"NO_TLSv1.2", SSL_OP_NO_TLSv1_2},
-#endif
-        {"PKCS1_CHECK_1", SSL_OP_PKCS1_CHECK_1},
-        {"PKCS1_CHECK_2", SSL_OP_PKCS1_CHECK_2},
-        {"NETSCAPE_CA_DN_BUG", SSL_OP_NETSCAPE_CA_DN_BUG},
-        {"NETSCAPE_DEMO_CIPHER_CHANGE_BUG",
-            SSL_OP_NETSCAPE_DEMO_CIPHER_CHANGE_BUG},
-#ifdef SSL_OP_CRYPTOPRO_TLSEXT_BUG
-        {"CRYPTOPRO_TLSEXT_BUG", SSL_OP_CRYPTOPRO_TLSEXT_BUG},
-#endif
-        {NULL, 0}
-    }, *option;
+    SSL_OPTION *option;
 
-    for(option=ssl_opts; option->name; ++option)
+    for(option=(SSL_OPTION *)ssl_opts; option->name; ++option)
         if(!strcasecmp(option->name, arg))
             return option->value;
     return 0; /* FAILED */
 }
 
+NOEXPORT void print_ssl_options(void) {
+    SSL_OPTION *option;
+
+    s_log(LOG_NOTICE, " ");
+    s_log(LOG_NOTICE, "Supported SSL options:");
+    for(option=(SSL_OPTION *)ssl_opts; option->name; ++option)
+        s_log(LOG_NOTICE, "options = %s", option->name);
+}
+
 /**************************************** socket options */
 
 static int on=1;
