@@ -40,26 +40,28 @@
 
 int main(int argc, char *argv[]) {
     static struct WSAData wsa_state;
-    char *c, stunnel_exe_path[MAX_PATH];
+    TCHAR *c, stunnel_exe_path[MAX_PATH];
 
     /* set current working directory and engine path */
     GetModuleFileName(0, stunnel_exe_path, MAX_PATH);
-    c=strrchr(stunnel_exe_path, '\\'); /* last backslash */
+    c=_tcsrchr(stunnel_exe_path, TEXT('\\')); /* last backslash */
     if(c) /* found */
-        c[1]='\0'; /* truncate program name */
+        c[1]=TEXT('\0'); /* truncate program name */
 #ifndef _WIN32_WCE
     if(!SetCurrentDirectory(stunnel_exe_path)) {
-        fprintf(stderr, "Cannot set directory to %s", stunnel_exe_path);
+        /* log to stderr, as s_log() is not initialized */
+        _ftprintf(stderr, TEXT("Cannot set directory to %s"),
+            stunnel_exe_path);
         return 1;
     }
 #endif
-    _putenv_s("OPENSSL_ENGINES", stunnel_exe_path);
+    _tputenv_s(TEXT("OPENSSL_ENGINES"), stunnel_exe_path);
 
     str_init(); /* initialize per-thread string management */
     if(WSAStartup(MAKEWORD(1, 1), &wsa_state))
         return 1;
     resolver_init();
-    main_initialize();
+    main_init();
     if(!main_configure(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL))
         daemon_loop();
     main_cleanup();
@@ -68,7 +70,7 @@ int main(int argc, char *argv[]) {
 
 /**************************************** options callbacks */
 
-void ui_new_config(void) {
+void ui_config_reloaded(void) {
     /* no action */
 }
 
@@ -94,12 +96,8 @@ void ui_clients(const int num) {
 
 /**************************************** s_log callbacks */
 
-void message_box(const LPSTR text, const UINT type) {
-    LPTSTR tstr;
-
-    tstr=str2tstr(text);
-    MessageBox(NULL, tstr, TEXT("stunnel"), type);
-    str_free(tstr);
+void message_box(LPCTSTR text, const UINT type) {
+    MessageBox(NULL, text, TEXT("stunnel"), type);
 }
 
 void ui_new_log(const char *line) {
